@@ -38,6 +38,12 @@ For more details, see individual docstrings.
 """
 
 from __future__ import annotations
+from itertools import zip_longest
+from copy import deepcopy
+from enum import Enum
+from functools import cached_property
+import warnings
+from collections import abc
 from typing import TYPE_CHECKING
 from typing import (
     Tuple,
@@ -49,12 +55,9 @@ from typing import (
     Union,
     TypeVar,
     Type,
+    Iterator,
 )
 
-from itertools import zip_longest
-from copy import deepcopy
-from enum import Enum
-
 from attr import define, field, Factory, cmp_using, setters  # type: ignore
 import numpy as np
 import numpy.typing as npt
@@ -64,7 +67,6 @@ from typicle import Types
 from typicle.convert import cast_array
 
 from ._base import ParticleBase, AdjacencyBase, MaskBase, ArrayBase
-from .calculate import eta, phi
 
 
 ###########################################
@@ -72,6 +74,7 @@ from .calculate import eta, phi
 ###########################################
 _types = Types()
 DoubleVector = npt.NDArray[np.float64]
+ComplexVector = npt.NDArray[np.complex128]
 BoolVector = npt.NDArray[np.bool_]
 IntVector = npt.NDArray[np.int32]
 HalfIntVector = npt.NDArray[np.int16]
@@ -79,6 +82,8 @@ ObjVector = npt.NDArray[np.object_]
 AnyVector = npt.NDArray[Any]
 DataType = TypeVar("DataType", bound=ArrayBase)
 
+DHUGE = np.finfo(np.dtype("<f8")).max * 0.1
+
 
 def _is_np_structured(array: AnyVector) -> bool:
     return array.dtype.names is not None
@@ -178,7 +183,7 @@ def _mask_dict_convert(masks: _IN_MASK_DICT) -> _MASK_DICT:
 
 
 @define
-class MaskGroup(MaskBase):
+class MaskGroup(MaskBase, abc.MutableMapping[str, MaskBase]):
     """Data structure to compose groups of masks over particle arrays.
     Can be nested to form complex hierarchies.
 
@@ -207,9 +212,12 @@ class MaskGroup(MaskBase):
         return cls(dict(map(lambda name: (name, arr[name]), arr.dtype.names)))
 
     def __repr__(self) -> str:
-        keys = ", ".join(self.names)
+        keys = ", ".join(map(lambda name: '"' + name + '"', self.names))
         return f"MaskGroup(mask_arrays=[{keys}], agg_op={self.agg_op.name})"
 
+    def __iter__(self) -> Iterator[str]:
+        return iter(self._mask_arrays)
+
     def __getitem__(self, key) -> Union[MaskArray, "MaskGroup"]:
         if not isinstance(key, str):
             return self.__class__(
@@ -232,7 +240,7 @@ class MaskGroup(MaskBase):
         self._mask_arrays.update({key: mask})
 
     def __len__(self) -> int:
-        return len(self.data)
+        return len(self.names)
 
     def __delitem__(self, key) -> None:
         """Remove a MaskArray from the group, using given key."""
@@ -471,19 +479,22 @@ class PdgArray(ArrayBase):
 ########################################
 @define
 class MomentumArray(ArrayBase):
-    """Returns data structure containing four-momentum of particle
-    list.
+    """Data structure containing four-momentum of particle list.
 
     Attributes
     ----------
-    data : ndarray
+    data : np.ndarray[double]
         Structured array containing four momenta.
-    pt : ndarray
+    pt : np.ndarray[double]
         Transverse component of particle momenta.
-    eta : ndarray
+    eta : np.ndarray[double]
         Pseudorapidity component of particle momenta.
-    phi : ndarray
+    phi : np.ndarray[double]
         Azimuthal component of particle momenta.
+    theta : np.ndarray[double]
+        Angular displacement from beam axis.
+    mass : np.ndarray[double]
+        Mass of the particles
     """
 
     data: AnyVector = array_field("pmu")
@@ -502,46 +513,96 @@ class MomentumArray(ArrayBase):
     def copy(self) -> "MomentumArray":
         return deepcopy(self)
 
-    @property
-    def _vector(self):
-        from vector import MomentumNumpy4D
+    @cached_property
+    def _xy_pol(self) -> ComplexVector:
+        return self.data["x"] + 1.0j * self.data["y"]  # type: ignore
 
-        dtype = deepcopy(self.data.dtype)
-        dtype.names = ("x", "y", "z", "t")
-        vec = self.data.view(dtype).view(MomentumNumpy4D)
-        return vec
+    @cached_property
+    def _zt_pol(self) -> ComplexVector:
+        return self.data["z"] + 1.0j * self.pt  # type: ignore
 
-    @property
+    @cached_property
+    def _spatial_mag(self) -> DoubleVector:
+        return np.abs(self._zt_pol)
+
+    @cached_property
     def pt(self) -> DoubleVector:
         """Momentum component transverse to the beam-axis."""
-        return self._vector.pt  # type: ignore
+        return np.abs(self._xy_pol)
 
-    @property
+    @cached_property
     def eta(self) -> DoubleVector:
-        """Pseudorapidity of particles."""
-        return eta(self)
+        """Pseudorapidity of particles.
 
-    @property
+        Notes
+        -----
+        Infinite values for particles travelling parallel to beam axis.
+        """
+        with warnings.catch_warnings():
+            warnings.simplefilter("ignore")
+            arr = np.arctanh(self.data["z"] / self._spatial_mag)
+        return arr  # type: ignore
+
+    @cached_property
+    def rapidity(self) -> DoubleVector:
+        """Rapidity of particles."""
+        e, z = self.data["e"], self.data["z"]
+        with warnings.catch_warnings():
+            warnings.simplefilter("ignore")
+            rap = 0.5 * np.log((e + z) / (e - z))  # type: ignore
+        return rap  # type: ignore
+
+    @cached_property
     def phi(self) -> DoubleVector:
-        """Azimuthal angular displacement of particles about the
-        beam-axis.
-        """
-        return phi(self)
+        """Azimuthal angular displacement of particles about beam-axis."""
+        return np.angle(self._xy_pol)
 
-    @property
+    @cached_property
     def theta(self) -> DoubleVector:
-        """Spherical angular displacement of particles from the positive
-        beam-axis.
-        """
-        return self._vector.theta  # type: ignore
+        """Angular displacement of particles from positive beam-axis."""
+        return np.angle(self._zt_pol)
 
-    @property
+    @cached_property
     def mass(self) -> DoubleVector:
-        """Mass of the particles."""
-        return self._vector.mass  # type: ignore
+        """Mass of particles."""
+        e: DoubleVector = self.data["e"]
+        p = self._spatial_mag
+        sq_diff = e * e - p * p
+        sign = np.sign(sq_diff)
+        return sign * np.sqrt(np.abs(sq_diff))  # type: ignore
 
-    def delta_R(self, other_pmu: "MomentumArray") -> DoubleVector:
-        return self._vector.deltaR(other_pmu._vector)  # type: ignore
+    def delta_R(self, other: "MomentumArray") -> DoubleVector:
+        """Calculates the Euclidean inter-particle distances in the
+        eta-phi plane between this set of particles and a provided
+        'other' set. Produces a mxn matrix, where m is number of
+        particles in this MomentumArray, and n is the number of
+        particles in other.
+
+        Parameters
+        ----------
+        other : MomentumArray
+            Four-momenta of the particle set to compute delta_R against.
+
+        Returns
+        -------
+        delta_R_matrix : np.ndarray[double]
+            Matrix representing the Euclidean distance between the two
+            sets of particles in the eta-phi plane. Rows represent
+            particles in this particle set, and columns particles in
+            the other set.
+
+        Notes
+        -----
+        Infinite values may be encountered if comparing with particles
+        not present on the eta-phi plane, __ie.__ travelling parallel to
+        the beam axis.
+        """
+        with warnings.catch_warnings():
+            warnings.simplefilter("ignore")
+            deta = self.eta[:, np.newaxis] - other.eta
+            deta = np.nan_to_num(deta, nan=0.0, posinf=np.inf, neginf=-np.inf)
+        dphi = np.angle(self._xy_pol[:, np.newaxis] * other._xy_pol.conj())
+        return np.hypot(deta, dphi)
 
 
 #################
